/*
 *  acm : an aerial combat simulator for X
 *  Copyright (C) 1991-1994  Riley Rainey
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; version 2 dated June, 1991.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program;  if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave., Cambridge, MA 02139, USA.
 */

#include <pm.h>
#ifndef WIN32
#include <X11/Xos.h>
#endif

static char *files[] =
{
	"engine.au",
	"crash.au",
	"gear_up.au",
	"gear_dn.au",
	"missile.au",
	"cannon.au",
	"crash.au",
	"screetch.au",
	"stall.au",
	"rwr.au",
	"apglock.au"
};

static char *paths[NUM_SOUNDS];

#if defined(NETAUDIO)

static int readPipe=-1;
static int writePipe=-1;

static AuServer * aserver;
static AuBucketID sound[NUM_SOUNDS];
static AuFlowID flow[NUM_SOUNDS] = {0};
static AuDeviceID audioOutput;

//#define TIMING 1
#ifdef TIMING
#include <sys/time.h>
#include <time.h>
#endif //TIMING

#include <sys/types.h>
#include <sys/wait.h>

struct packet_t
{
	enum { PLAY, STOP } command;
	int sound;
	int repeat;
};

static int
local_initializeAudio( const char *display, const char *auserver);

static void
local_shutdownAudio();

static void
local_playSound(int id, int repeat);

static void
local_stopSound(int id);

static void handle_pipe(int i)
{
	/* hopefully this was because we wrote to the audio pipe
	 * and the audio daemon has died.
	 */
	writePipe=-1;
}

static void handle_child(int i)
{
	int status;
	/* Audio daemon must have gone away.
	 */
	writePipe=-1;
	waitpid(-1, &status, WNOHANG);
}

int
initializeAudio(c, v, display, auserver)
craft    *c;
viewer   *v;
char     *display;
char     *auserver;
{
	int filedesc[2];
	int pid, i;
	struct packet_t p;
	int result;

	fprintf(stderr, "About to start sound daemon.\n");
	if(pipe(filedesc)==-1) {
		perror("Error creating pipe");
		return -1;
	}
	signal(SIGPIPE, handle_pipe);
	signal(SIGCHLD, handle_child);
	pid = fork();
	if(pid == -1) {
		perror("Fork error");
		v->aserver = NULL;
		signal(SIGPIPE, SIG_DFL);
		signal(SIGCHLD, SIG_DFL);
		return -1;
	}
	if( pid ) {
		close(filedesc[0]);
		writePipe = filedesc[1];
		v->aserver = (AuServer*)pid;
		for( i =0; i<NUM_SOUNDS; i++){
			v->flow[i] = 0;
		}
		return 0;
	}
	signal(SIGPIPE, SIG_DFL);
	signal(SIGCHLD, SIG_DFL);

	close(filedesc[1]);
	readPipe = filedesc[0];

	if( local_initializeAudio( display, auserver) ){
		close(readPipe);
		fprintf(stderr, "Error setting up audio\n");
		exit(1);
	}

	result = read( readPipe, &p, sizeof(struct packet_t));
	while( result == sizeof(struct packet_t)) {
		if(p.command == PLAY)
			local_playSound( p.sound, p.repeat);
		else if(p.command == STOP)
			local_stopSound( p.sound);
		else
			fprintf(stderr, "Audio Daemon received a bad command\n");

		result = read( readPipe, &p, sizeof(struct packet_t));
	}
	if(result == -1)
		perror("Error reading pipe");
	local_shutdownAudio();

	close(readPipe);
	fprintf(stderr, "Audio daemon exiting.\n");
	exit(1);
}

static void
writeCommand( struct packet_t *p)
{
	int result;
	int loc=0;
	do
	{
		result = write(writePipe, p+loc, sizeof(struct packet_t)-loc);
		if(result == -1 || !result){
			if(result == -1)
				perror("Error writing to daemon");
			writePipe = -1;
			return;
		}
		loc += result;
	}while(loc < sizeof(struct packet_t));
}

void
shutdownAudio(c, v)
craft    *c;
viewer   *v;
{
	// this should close the daemon down
	close(writePipe);
	v->aserver = (AuServer *) NULL;
	wait(NULL);
}

void
playSound(c, id)
craft    *c;
int       id;
{
	struct packet_t p;
#ifdef TIMING
	struct timeval start,stop;
	double total;
	gettimeofday(&start, NULL);
#endif //TIMING
	if (c->type == CT_PLANE) {
	  if(writePipe != -1) {
	    p.command = PLAY;
	    p.sound = id;
	    p.repeat = 0;
	    writeCommand(&p);
	    if(c->vl)
	      c->vl->flow[id]=1;
	  }
	}

#ifdef TIMING
	gettimeofday(&stop, NULL);
	total = stop.tv_sec - start.tv_sec + (stop.tv_usec - start.tv_usec)/1000000.0;
	fprintf(stderr, "%s\t%d\t %f\n", __FUNCTION__, id, total);
#endif //TIMING
}

void
playContinuousSound(c, id)
craft    *c;
int       id;
{
	struct packet_t p;
#ifdef TIMING
	struct timeval start,stop;
	double total;
	gettimeofday(&start, NULL);
#endif //TIMING

	if ((c->type = CT_PLANE)) {
	  if(writePipe != -1) {
	    p.command = PLAY;
	    p.sound = id;
	    p.repeat = 10;
	    writeCommand(&p);
	    if(c->vl)
	      c->vl->flow[id]=1;
	  }
	}


#ifdef TIMING
	gettimeofday(&stop, NULL);
	total = stop.tv_sec - start.tv_sec + (stop.tv_usec - start.tv_usec)/1000000.0;
	fprintf(stderr, "%s\t%d\t %f\n", __FUNCTION__, id, total);
#endif //TIMING
}

void
stopSound(c, id)
craft    *c;
int       id;
{
	struct packet_t p;
#ifdef TIMING
	struct timeval start,stop;
	double total;
	gettimeofday(&start, NULL);
#endif //TIMING


	if ((c->type = CT_PLANE)) {
	  if(writePipe != -1) {
	    p.command = STOP;
	    p.sound = id;
	    p.repeat = 0;
	    writeCommand(&p);
	    if(c->vl)
	      c->vl->flow[id]=0;
	  }
	}

#ifdef TIMING
	gettimeofday(&stop, NULL);
	total = stop.tv_sec - start.tv_sec + (stop.tv_usec - start.tv_usec)/1000000.0;
	fprintf(stderr, "%s\t%d\t %f\n", __FUNCTION__, id, total);
#endif //TIMING
}

static void done_callback();

static int
local_initializeAudio( const char * display, const char * auserver)
{
	int       i, n;
	AuBucketAttributes *ret_attr;
	AuDeviceAttributes *dev;
	AuStatus  status;
	char * the_server;
	char * dot_position;

/*
 *  If the X display specification has a screen field, remove it so that we
 *  can use it as an audio server name.
 */

	the_server = strdup(auserver);
	dot_position = strchr(the_server, '.');
	if (dot_position) {
	  *dot_position = '\0';
	}

/*
 *  Lazy initialization of sound paths array
 */

	if (!paths[0]) {
		for (i = 0; i < NUM_SOUNDS; ++i) {
			paths[i] = strdup(acm_find_file(files[i]));
		}
	}

/*
 *  Open the audio server and send it our sounds
 */

	if ((aserver = AuOpenServer(the_server, 0, (char *) NULL,
							0, (char *) NULL, (char **) NULL)) == NULL) {
		aserver = (AuServer *) NULL;
		return -1;
	}

	for (i = 0; i < NUM_SOUNDS; ++i) {
		if ((sound[i] = AuSoundCreateBucketFromFile(aserver,
						 paths[i], AuAccessListMask | AuAccessImportMask,
										&ret_attr, &status)) == AuNone) {
			printf("Unable to create sound bucket: %s\n",
				   paths[i]);
			AuCloseServer(aserver);
			aserver = (AuServer *) NULL;
			return -1;
		}
		flow[i] = 0;
	}

/*
 *  Select an output device
 */

	audioOutput = AuNone;
	for (i = 0; i < AuServerNumDevices(aserver); ++i) {
		dev = AuServerDevice(aserver, i);
		if (AuDeviceKind(dev) == AuComponentKindPhysicalOutput &&
			AuDeviceNumTracks(dev) == 1) {
			audioOutput = AuDeviceIdentifier(dev);
			break;
		}
	}

	return 0;
}

static void
local_shutdownAudio()
{
	int       i;
	AuStatus  status;

	if (aserver) {
		for (i = 0; i < NUM_SOUNDS; ++i) {
			AuDestroyBucket(aserver, sound[i], &status);
		}
		AuCloseServer(aserver);
		aserver = (AuServer *) NULL;
		for (i = 0; i < NUM_SOUNDS; ++i) {
			free (paths[i]);
		}
	}
}

static void
local_playSound(id, repeat)
int       id, repeat;
{
	AuStatus  status;
	if (aserver) {
		AuSoundPlayFromBucket(aserver, sound[id],
			audioOutput, AuFixedPointFromSum(1, 0),
			done_callback, (AuPointer) NULL,
			repeat, &flow[id], (int *) NULL, (int *) NULL,
			&status);
	}
}

static void
local_stopSound(id)
int       id;
{
	AuStatus  status;

	if (aserver != (AuServer *) NULL && flow[id] != 0) {
		AuStopFlow(aserver, flow[id], &status);
		AuDestroyFlow(aserver, flow[id], &status);
		flow[id] = 0;
	}
}

static void
done_callback(audio, which, event, data)
AuServer *audio;
AuEventHandlerRec *which;
AuEvent  *event;
AuPointer data;
{
	printf("event type = %d\n", event->type);
}

#else							/* defined(NETAUDIO) */
#if defined (HPAUDIO)

#undef	duration				/* defined by pm.h as "radarMode" */

int
initializeAudio(c, v, display)
craft    *c;
viewer   *v;
char     *display;
{
	int       i, n;
	Audio    *conn;
	AudioAttributes sb_attr;
	long      status;

/*
 *  Lazy initialization of sound paths array
 */

	if (!paths[0]) {
		for (i = 0; i < NUM_SOUNDS; ++i) {
			paths[i] = strdup(acm_find_file(files[i]));
		}
	}

/*
 *  Open the audio server and send it our sounds
 */

	if ((v->aserver = AOpenAudio(display, &status)) == NULL) {
		printf("AOpenAudio: failed, status = %d\n", status);
		v->aserver = (Audio *) NULL;
		return -1;
	}

	for (i = 0; i < NUM_SOUNDS; ++i) {
		if ((v->sound[i] = ALoadAFile(v->aserver,
									  paths[i], AFFUnknown, 0, &sb_attr,
									  &status)) == NULL) {
			printf("Unable to create sound bucket: %s (status %d)\n",
				   paths[i], status);
			ACloseAudio(v->aserver, &status);
			v->aserver = (Audio *) NULL;
			return -1;
		}
		v->flow[i] = 0;
	}

	return 0;
}

void
shutdownAudio(c, v)
craft    *c;
viewer   *v;
{
	int       i;
	long      status;

	if (v->aserver) {
		for (i = 0; i < NUM_SOUNDS; ++i) {
			ADestroySBucket(v->aserver, v->sound[i], &status);
		}
		ACloseAudio(v->aserver, &status);
		v->aserver = (Audio *) NULL;
		for (i = 0; i < NUM_SOUNDS; ++i) {
			free(paths[i]);
		}
	}
}

void
playSound(c, id)
craft    *c;
int       id;
{
	viewer   *v;
	SBPlayParams pparams;
	AGainEntry gainEntry;
	long      status;

	gainEntry.u.o.out_ch = AOCTMono;
	gainEntry.gain = AUnityGain;
	gainEntry.u.o.out_dst = AODTMonoIntSpeaker;

	bzero(&pparams, sizeof(pparams));
	pparams.priority = APriorityNormal;
	pparams.gain_matrix.type = AGMTOutput;
	pparams.gain_matrix.num_entries = 1;
	pparams.gain_matrix.gain_entries = &gainEntry;
	pparams.play_volume = AUnityGain;
	pparams.pause_first = False;
	pparams.start_offset.type = ATTSamples;
	pparams.start_offset.u.samples = 0;
	pparams.duration.type = ATTFullLength;
	pparams.loop_count = 0;
	pparams.previous_transaction = 0;
	pparams.event_mask = 0;

	for (v = c->vl; v; v = v->next) {
		if (v->aserver) {
			v->flow[id] = APlaySBucket(v->aserver, v->sound[id],
									   &pparams, &status);
			if (v->flow[id] == NULL)
				printf("APlaySBucket: failed, status %d\n", status);
		}
	}
}

void
playContinuousSound(c, id)
craft    *c;
int       id;
{
	viewer   *v;
	AGainEntry gainEntry;
	SBPlayParams pparams;
	long      status;

	gainEntry.u.o.out_dst = AODTMonoIntSpeaker /* AODTMonoJack */ ;
	gainEntry.u.o.out_ch = AOCTMono;
	gainEntry.gain = AUnityGain;

	bzero(&pparams, sizeof(pparams));
	pparams.gain_matrix.type = AGMTOutput;
	pparams.gain_matrix.num_entries = 1;
	pparams.gain_matrix.gain_entries = &gainEntry;
	pparams.play_volume = AUnityGain;
	pparams.pause_first = False;
	pparams.start_offset.type = ATTSamples;
	pparams.start_offset.u.samples = 0;
	pparams.duration.type = ATTFullLength;
	pparams.loop_count = -1;	/* repeating... */
	pparams.priority = APriorityNormal;
	pparams.previous_transaction = 0;
	pparams.event_mask = 0;

	for (v = c->vl; v; v = v->next) {
		if (v->aserver) {
			v->flow[id] = APlaySBucket(v->aserver, v->sound[id],
									   &pparams, &status);
			if (v->flow[id] == NULL)
				printf("APlaySBucket: failed, status %d\n", status);
		}
	}
}

void
stopSound(c, id)
craft    *c;
int       id;
{
	viewer   *v;
	long      status;

	for (v = c->vl; v; v = v->next) {
		if (v->aserver != (Audio *) NULL && v->flow[id] != 0) {
			AStopAudio(v->aserver, v->flow[id], ASMEndLoop, NULL, &status);
			if (status)
				printf("AStopAudio: status %d\n", status);
			v->flow[id] = 0;
		}
	}
}

#else							/* defined(HPAUDIO) */

int
initializeAudio(craft * c, viewer * v, char *p)
{
	int       i;

	for (i = 0; i < NUM_SOUNDS; ++i) {
		v->flow[i] = 0;
	}
	return -2;
}

void
shutdownAudio(craft * c, viewer * v)
{
}

void
playSound(craft * c, int id)
{
}

void
playContinuousSound(craft * c, int id)
{
}

void
stopSound(craft * c, int id)
{
}

#endif							/* defined(HPAUDIO) */
#endif							/* defined(NETAUDIO) */

void
setBackgroundSound(craft * c, double dThtlPercent,
				   _BOOL bABFlag, double dDynamicPressure)
{
}
