/*
 *  acm : an aerial combat simulator for X
 *  	: Joystick support for the Linux Joystick driver v2.0.0
 *  	: documentation on the joystick driver is available in the
 *  	: linux kernel in linux/Documentation/input/joystick.txt
 *  	: and http://linuxconsole.sourceforge.net/input/joystick.html
 *  Copyright (C) 2003  David Fries
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; version 2 dated June, 1991.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program;  if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave., Cambridge, MA 02139, USA.
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>

static int fd = -1;
int LinuxJoystickPresent = 0;

void
LinuxSetJoystickPort(const char * name)
{
	fd = open(name, O_RDONLY|O_NONBLOCK);
	if(fd == -1)
	{
		fprintf(stderr, "Error opening joystick device ");
		perror(name);
		return;
	}
	LinuxJoystickPresent = 1;
}

void
LinuxCalibrateJoystick()
{
	fprintf(stderr, "Please run jscal to calibrate "
		"the Linux joystick driver.\n");
	fprintf(stderr,
		"http://linuxconsole.sourceforge.net/input/joystick.html\n");
}

#ifndef __linux__
#warning useful only on Linux
int
LinuxGetJoystickPosition( double *x1, double *y1,
	double *rudder, double *throttle, int *switches)
{
	return -1;
}	
#else
#include <linux/joystick.h>

/* x1 ailerons
 * y1 elevator
 * x2 rudder
 * y2 throttle
 * switches bit 0 guns, bit 1 select weapon
 */
int
LinuxGetJoystickPosition( double *x1, double *y1,
	double *rudder, double *throttle, int *switches)
{
	struct js_event jevent;
	#define AXIS_MAX 4
	// rudder input is -1 to 1 with -2 being no rudder, same for throttle
	static double axis[AXIS_MAX] = {0,0,-2,-2};
	/* If we get a button press event followed by a release event
	 * we still want to report that the button had been pressed.
	 * When we enter the routine release any pressed buttons that
	 * have release set on them, then reset the release.
	 * As we process the events a button press sets the buttons
	 * and unsets the release, while a button release sets the
	 * release.  This is so the calling program always sees a button
	 * press no matter what the interval, while it might not
	 * see that a button was released for a short amount of time.
	 */
	int previous;
	static int buttons = 0;
	static int release = 0;
	int result;
	int b;

	if(fd == -1)
		return -1;

	// reset any buttons that had been released, yet hadn't been reported
	if((buttons & 1) && (release & 1 ))
		buttons &= ~1;
	if((buttons & 2) && (release & 2 ))
		buttons &= ~2;
	release = 0;

	previous = buttons;

	while((result = read(fd, &jevent, sizeof(jevent)))!=-1)
	{
		switch(jevent.type & ~JS_EVENT_INIT)
		{
		case JS_EVENT_BUTTON:
			b = 1 << jevent.number;
			if(jevent.value) // press
			{
				buttons |= b;
				release &= ~b;
			}
			else // release
			{
				// we've arleady sent a press
				if(previous & b)
				{
					buttons &= ~b;
					release &= ~b;
					previous &= ~b;
				}
				else
				{
					release |= b;
				}
			}
			break;
		case JS_EVENT_AXIS:
			if(jevent.number < AXIS_MAX)
				axis[jevent.number] =
					-1 * (double)jevent.value/32768.0;
		}
	}
	*x1 = axis[0];
	*y1 = axis[1];
	*rudder = axis[2];
	*throttle = axis[3];
	*switches = buttons;

	return 0;
}
#endif
